import logging
import os
import re
from collections import OrderedDict as OrderedDictColl
from datetime import datetime
from os import path
from typing import List, Optional, Text, OrderedDict, Dict, Union, Tuple
from uuid import uuid4

from siena.shared.constants import (
    DEFAULT_DATA_PATH,
    DEFAULT_NLU_YAML_TAG,
    DEFAULT_NLU_YAML_VERSION,
    DEFAULT_VERSION_YAML_TAG,
    YAML_EXTENSIONS,
    FilePermission,
    Encoding,
    DEFAULT_NLU_INTENT_TAG,
    DEFAULT_NLU_EXAMPLES_TAG,
    TermColor,
    DEFAULT_CASE_SENSITIVE_MODE,
)
from siena.shared.exceptions.io import (
    YAMLFormatException,
    NLUFileNotFoundException,
    InvalidFileExtensionException,
    FileSizeInspectingException,
)
from rasa.shared.data import get_data_files, is_nlu_file
from ruamel import yaml as yaml
from ruamel.yaml.error import YAMLError

logger = logging.getLogger(__name__)


def _get_dir_file_list(
        dir_path: Text = DEFAULT_DATA_PATH,
        file_suffixes: List = YAML_EXTENSIONS
) -> Optional[List]:
    """
    Returns the list files available in a given
    directory path and its subdirectories. able
    to check extensions if the extensions are
    provided as a list

    Args:
        dir_path: path where to read the list of directories
        file_suffixes: list of file extensions to look for. if
            not provided, the method will output all the files
            available in the given destination directory

    Returns:
        list of files available, or None
    """

    file_paths = list()

    for (dir_path, dir_names, file_names) in os.walk(dir_path):
        if file_suffixes:
            file_paths += [path.join(dir_path, file) for file in file_names if str(file).endswith(tuple(file_suffixes))]
        else:
            file_paths += [path.join(dir_path, file) for file in file_names]
    return file_paths


def get_all_existing_file_list(
        dir_path: Text = DEFAULT_DATA_PATH,
) -> Optional[List]:
    """
    Returns all files and directories existing
    in the specified destination directory and
    subdirectories

    Args:
        dir_path: destination directory path where
            to get the list of files and directories

    Returns:
        list of files and directories available, or
            None
    """

    file_list_all = list()

    for (dir_path, dir_names, file_names) in os.walk(dir_path):
        file_list_all += [dir_ for dir_ in dir_names]
        file_list_all += [file for file in file_names]

    return file_list_all


def get_existing_toplevel_file_list(
        dir_path: Text = DEFAULT_DATA_PATH,
        exclude: List = None
) -> Optional[List]:
    """
    Returns the list files available in a given
    directory path but not subdirectories. able
    to ignore files if the provided as a list under
    exclude argument

    Args:
        dir_path: path where to read the list of directories
        exclude: list of file names to ignore

    Returns:
        list of files available except the files mentioned in
            exclude list, or None
    """

    files = os.listdir(dir_path)
    if exclude:
        files = list(set(files).difference(set(exclude)))
    return files


def get_timestamp_str(sep: Text = "-", uuid: bool = False) -> Text:
    """
    Generates a timestamped string and attaches a unique
    ID using UUID if specified and returns it

    Args:
        sep: seperator to separate date, time and uuid
        uuid: if True, a unique UUID will be attached
            at the end of the timestamped string
            generated

    Returns:
        timestamped string
    """
    if uuid:
        return datetime.now().strftime('%Y%m%d' + sep + '%H%M%S' + sep) \
               + str(uuid4())
    else:
        return datetime.now().strftime('%Y%m%d' + sep + '%H%M%S')


def dir_exists(dir_path: Text = None) -> bool:
    """
    Checks if the specified directory path exists

    Args:
        dir_path: path of the directory to check
            the existence

    Returns:
        True if path exists, else False
    """
    return path.exists(dir_path) and path.isdir(dir_path)


def file_exists(file_path: Text = None) -> bool:
    """
    Checks if the specified file path exists

    Args:
        file_path: path of the file to check
            the existence

    Returns:
        True if path exists, else False
    """
    return path.exists(file_path) and path.isfile(file_path)


def file_size(file_path: Union[Text, List], reverse: bool = False) -> Union[Dict, Tuple[float, Text]]:
    try:
        if isinstance(file_path, Text):
            size_in_bytes = os.path.getsize(file_path)
            if size_in_bytes >= 1000000000:
                size = size_in_bytes / 1000000000
                units = "GB"
            elif size_in_bytes >= 1000000:
                size = size_in_bytes / 1000000
                units = "MB"
            elif size_in_bytes >= 1000:
                size = size_in_bytes / 1000
                units = "KB"
            else:
                size = size_in_bytes
                units = "Bytes"
            return size, units
        else:
            if reverse:
                file_path.reverse()
            file_sizes_dict = {file: file_size(file_path=file) for file in file_path}
            return file_sizes_dict
    except Exception as e:
        raise FileSizeInspectingException(e)


def read_yaml_file(
        yaml_file: Text = None,
        encoding: Text = Encoding.UTF8,
        mode: Text = FilePermission.READ,
        yaml_version: Text = DEFAULT_NLU_YAML_VERSION,
        version_check: bool = True,
) -> Optional[OrderedDict]:
    """
    Reads the content of a specified YAML file.

    Args:
        yaml_file: YAML file to read the content from
        encoding: encoding of the specified file
        mode: read mode
        yaml_version: version of the YAML file, 2.0 is
            preferred
        version_check: True if required to check the
            version, else False

    Returns:
        YAML file content as an ordered dictionary, or
            None
    """
    if not yaml_file:
        return OrderedDictColl()

    with open(file=yaml_file, encoding=encoding, mode=mode) as file_stream:
        try:
            yaml_content = yaml.round_trip_load(file_stream, preserve_quotes=False)
        except YAMLError as e:
            raise YAMLFormatException(e)

    if not yaml_content:
        return None

    if version_check:
        if DEFAULT_VERSION_YAML_TAG not in yaml_content:
            logger.debug("The YAML file is not properly versioned. the expected version is 2.0")
        else:
            if yaml_content[DEFAULT_VERSION_YAML_TAG] != yaml_version:
                logger.debug("The YAML file does not contain the expected rasa YAML version. the expected version is "
                             "2.0")

    return yaml_content


def find_yaml_collection(
        yaml_content: OrderedDict = None,
        yaml_collection_tag: Text = DEFAULT_NLU_YAML_TAG,
) -> OrderedDict:
    """
    Given the content of a YAML file, finds and returns
    the specified YAML collection.

    Args:
        yaml_content: content of a YAML file as an ordered
            dictionary
        yaml_collection_tag: tag of the collection to be
            read

    Returns:
        The YAML collection extracted out of the YAML content
            as an ordered dictionary
    """
    if yaml_content is None:
        return OrderedDictColl()
    if yaml_collection_tag not in yaml_content:
        return OrderedDictColl()
    return yaml_content[yaml_collection_tag]


def _sanitize_nlu_data(
        collection_content: Optional[OrderedDict],
        case_sensitive: bool = DEFAULT_CASE_SENSITIVE_MODE
) -> Optional[Dict]:
    """
    Receives a collection of already retrieved
    RASA NLU examples and cleans it. Returns a
    sanitized non-empty collection of nlu example
    sentences.

    Args:
        collection_content: input data as an ordered
            dictionary
        case_sensitive:  if True, reads all data as
            lowercase, or else preserves the case and
            reads the files as is

    Returns:
        sanitized NLU data as a dictionary which
            has the list of examples under 'examples'
            under the key 'intent'
    """

    if not collection_content:
        return None

    nlu_examples = {
        intent[DEFAULT_NLU_INTENT_TAG]: [
            str.strip(instance) for instance in
            (str.split(
                re.sub("- ", "", intent[DEFAULT_NLU_EXAMPLES_TAG]),
                "\n"
            ))
            if str.strip(instance) not in ['']
        ]
        for intent in collection_content
        if len(intent) == 2
    }

    if not case_sensitive:
        nlu_examples = {
            intent: [str.lower(example) for example in examples]
            for intent, examples
            in nlu_examples.items()
        }
    return nlu_examples


def get_unique_list(list_of_data: Optional[List] = None) -> Optional[List]:
    """
    Converts a given list to a list that only contains
    unique elements. Eliminates duplicate elements by
    converting to a set and back to a list.

    Args:
        list_of_data: list of elements

    Returns:
        list containing only unique elements, or None
    """

    if not list_of_data:
        return None
    return list(set(list_of_data))


def get_rasa_testing_data(
        testing_data_dir: Text = DEFAULT_DATA_PATH,
        file_ext: List = YAML_EXTENSIONS,
        case_sensitive: bool = DEFAULT_CASE_SENSITIVE_MODE
) -> Optional[Dict]:
    """
    Reads testing data from rasa NLU testing
    data YAML files and returns a single list
    of sanitized testing data examples

    Args:
        testing_data_dir: path of the directory where to look
            for RASA testing data
        file_ext: extension of the testing data files to look
            for specified as a list. default is YAML or YML
        case_sensitive: if True, reads all data as lowercase, or
            else preserves the case and reads the files as is

    Returns:
        testing data as a dictionary with 'intent' and list of
            'examples' under each 'intent' key

    Raises:
        InvalidFileExtensionException: when the extension of the
            provided file is not a valid YAML file
    """

    try:
        testing_data = dict()
        logger.debug("Initializing testing data")

        if file_exists(file_path=testing_data_dir):
            file_extension = os.path.splitext(testing_data_dir)[-1]
            if file_extension in YAML_EXTENSIONS:
                testing_data_files = [testing_data_dir]
            else:
                raise InvalidFileExtensionException("The specified file is not a valid "
                                                    "YAML file")
        else:
            testing_data_files = _get_dir_file_list(dir_path=testing_data_dir, file_suffixes=file_ext)
        logger.info(f"{len(testing_data_files)} YAML files were found in the testing data directory.")

        if testing_data_files:
            logger.debug(f"List of files found: {', '.join([os.path.split(fp)[-1] for fp in testing_data_files])}")

        for file in testing_data_files:
            try:
                nlu_data = read_yaml_file(
                    yaml_file=file,
                    version_check=True
                )
            except FileNotFoundError:
                raise NLUFileNotFoundException("Could not locate one or more NLU testing data YAML files")
            nlu_content = find_yaml_collection(
                yaml_content=nlu_data,
                yaml_collection_tag=DEFAULT_NLU_YAML_TAG
            )
            cleaned_nlu_content = _sanitize_nlu_data(
                collection_content=nlu_content,
                case_sensitive=case_sensitive
            )

            if not cleaned_nlu_content:
                continue

            for intent, examples in cleaned_nlu_content.items():
                if intent in testing_data:
                    testing_data[intent] += examples
                else:
                    testing_data[intent] = examples

        return testing_data
    except NLUFileNotFoundException as e:
        logger.error(e)
    except YAMLFormatException as e:
        logger.error(e)
    except Exception as e:
        logger.error(e)


def set_cli_color(text_content: Text = None, color: str = TermColor.NONE_C):
    """
    wraps inbound string instances with ASCII
    color codes and returns a colored string
    that can be printed on the terminals

    Args:
        text_content: string instance to wrap
        color: color code to be visible on
            terminals

    Returns:
        wrapped string with the specified color
            that is ready to be printed on a
            terminal
    """
    return color + str(text_content) + TermColor.END_C


def series_to_json_serializable(series: Union[Dict, List]):
    if isinstance(series, dict):
        return {k: float(v) for k, v in series.items()}
    elif isinstance(series, list):
        return [float(v) for v in series]


def testing_data_dir_exists(dir_path: Text) -> bool:
    if not os.path.exists(dir_path):
        return False

    data_list = get_data_files(paths=[dir_path], filter_predicate=is_nlu_file)
    if len(data_list) == 0:
        return False

    return True


import logging
import os
from typing import Dict

from waitress import serve as waitress_serve

from siena.core.actions import init_project
from siena.server import create_app
from siena.shared.constants import (
    ServerEnv,
    SIENAConfig,
    DEFAULT_SERVER_HOST,
    DEFAULT_SERVER_PORT,
    ASCII_LOGO,
)
from siena.shared.exceptions.base import (
    SIENABaseException,
)
from siena.shared.exceptions.core import (
    SIENACoreException,
)
from siena.shared.exceptions.server import (
    SIENAServerException,
)

logger = logging.getLogger(__name__)


class SIENAServer:
    def __init__(
            self,
            configs: Dict = None,
            debug_mode: bool = False,
    ):
        self.configs = configs
        self.debug_mode = debug_mode
        self.host = configs[SIENAConfig.SUB_KEY_SERVER_HOST] or DEFAULT_SERVER_HOST
        self.port = configs[SIENAConfig.SUB_KEY_SERVER_PORT] or DEFAULT_SERVER_PORT

        # initialize project
        init_project()

    def run(self) -> None:
        logger.info(f"Starting SIENA server at http://{self.host}:{self.port}")
        try:
            if self.debug_mode:
                logger.warning("Deploying SIENA Server in development mode...")
                os.environ["APP_ENV"] = ServerEnv.DEV
                app = create_app()
                app.run(
                    host=self.host,
                    port=self.port,
                    debug=self.debug_mode
                )
            else:
                logger.info("Deploying SIENA Server in production mode...")
                print(ASCII_LOGO)
                waitress_serve(create_app(), host=self.host, port=self.port)

                # # Run as a shell command if required
                # import subprocess
                # subprocess.run(["waitress-serve", f"--host={self.host}",
                #                 f"--port={self.port}", "siena.server.siena_server:run"])

        except SIENACoreException:
            logger.exception(f"Core Exception")
        except SIENAServerException:
            logger.exception(f"Server Exception")
        except SIENABaseException:
            logger.exception(f"Base Exception")
        except KeyboardInterrupt:
            logger.info(f"Gracefully terminating SIENA Server...")
            exit(1)
        except OSError:
            logger.exception(f"Possible permission exception while starting the SIENA Server")
        except Exception as e:
            logger.exception(f"Base Exception: Broad. more info: {e}")
        return



import logging
import math
import os
import string
import nltk
import numpy as np
import pandas as pd
from nltk.stem.porter import *
from ruamel.yaml import YAML

from siena.shared.constants import (
    SUFFIXES_RAW,
    VOWELS_MAPPER,
    COLUMN_NAME_BASE_WORD,
    COLUMN_NAME_COUNT,
    COLUMN_NAME_ENTITY,
    SIENA_KNOWLEDGE_BASE_PATH,
    FilePermission,
    Encoding,
    SIENA_TEMP_KNOWLEDGE_BASE_PATH,
    NEW_LINE_TAG,
    NLTK_WORDNET,
    NLTK_PUNKT,
    EN_ALPHABET,
)
logger = logging.getLogger(__name__)
en_stemmer = PorterStemmer()
nltk.download(NLTK_WORDNET, quiet=True)
nltk.download(NLTK_PUNKT, quiet=True)
yml = YAML()
yml.indent(mapping=2, sequence=4, offset=2)
yml.preserve_quotes = True
yml.explicit_start = False

vowels_mapper = VOWELS_MAPPER
suffixes_raw = SUFFIXES_RAW
# encoder
key_list = list(vowels_mapper.keys())
# changes the order to increase the accuracy (lengthy letters should map first)
key_list_rev = list(reversed(key_list))


def si_vowels_encoder(vowels_mapper_, key_list_rev_, word):
    for letter in key_list_rev_:
        vowel = vowels_mapper_[letter]
        word = word.replace(letter, vowel)
        word = word.replace('අ්', '')
    return word


# decoder
vowels_mapper_inversed = {}
for key, value in vowels_mapper.items():
    vowels_mapper_inversed[value] = key

key_list_inversed = list(vowels_mapper_inversed.keys())
# changes the order to increase the accuracy (lengthy letters should map first)
key_list_rev_inversed = list(reversed(key_list_inversed))


def si_vowels_decoder(vowels_mapper_inversed_, key_list_rev_inversed_, base_word_encoded):
    for letter in key_list_rev_inversed_:
        vowel = vowels_mapper_inversed_[letter]
        if letter in base_word_encoded:
            base_word_encoded = base_word_encoded.replace(letter, vowel)
            # base_word_encoded = base_word_encoded.replace('අ්','')
    return base_word_encoded


# suffixes
suffixes_unique = suffixes_raw.split(NEW_LINE_TAG)
suffixes_unique = list(set([letter_processed.strip() for letter_processed in suffixes_unique]))

suffixes_processed = [si_vowels_encoder(vowels_mapper, key_list_rev, word) for word in suffixes_unique]


def si_stemmer_sentence_custom(word: str) -> str:
    res = []
    # Initialising string
    ini_string = si_vowels_encoder(vowels_mapper, key_list_rev, word)
    for suffix in suffixes_processed:
        if ini_string.endswith(suffix):
            res.append(ini_string[:-(len(suffix))])

        # printing result
    if len(res) > 0:
        result = min(res, key=len)
    else:
        result = word
    # decode base word
    result = si_vowels_decoder(vowels_mapper_inversed, key_list_rev_inversed, result)
    return result


def en_stemmer_sentence(text: str) -> str:
    tokenization = nltk.word_tokenize(text)
    en_stemmed = ""
    for w in tokenization:
        en_stemmed += " " + en_stemmer.stem(w)
    return en_stemmed


def base_form_convetor(selection: str) -> str:
    text = selection.translate(str.maketrans('', '', string.punctuation))
    en_stemmed = en_stemmer_sentence(text)
    si_stemmed = si_stemmer_sentence_custom(en_stemmed)
    return si_stemmed.strip()


# Similarity Algorithms 
alphabet = EN_ALPHABET
en_alphabet = alphabet.split()
si_alphabet = key_list
vecSpace = tuple(si_alphabet + en_alphabet)


def cosing_sim(word1: str, word2: str) -> float:
    words = [word1, word2]
    # counting chars
    result = []
    for word in words:
        word_vec = []
        for letter in vecSpace:
            word_vec.append(word.count(letter))
        squares = list(map(lambda x: pow(x, 2), word_vec))
        sum_of_sq = sum(squares)
        norm = math.sqrt(sum_of_sq)
        norm_vec = [i / norm for i in word_vec]
        result.append(norm_vec)

    result = np.array(result)
    df = pd.DataFrame()
    i = 0
    for singleVec_R in result:
        col = []
        for singleVec_C in result:
            col.append(np.sum(singleVec_R * singleVec_C))
        df[i] = col
        i += 1
    return float(df[0][1])


def generate_ngrams(s, n):
    tokens = list(s)
    ngrams = zip(*[tokens[i:] for i in range(n)])
    return ["".join(ngram) for ngram in ngrams]


def generate_ngrams_sent(sentence, n):
    sentence_split = sentence.split(" ")
    tokens = list(sentence_split)
    ngrams = zip(*[tokens[i:] for i in range(n)])
    return [" ".join(ngram) for ngram in ngrams]


def n_gram_similarity(word1, word2):
    word_1 = generate_ngrams(word1, 2)
    word_2 = generate_ngrams(word2, 2)
    count = 0
    ans = 0
    if len(word_1) > len(word_2):
        long_word = word_1
        short_word = word_2
    else:
        long_word = word_2
        short_word = word_1
    try:
        for element in short_word:
            if element in long_word:
                count += 1
        ans = count / (len(long_word))
    except Exception as e:
        logger.exception(f"Exception occurred. {e}")

    return ans


def similarity(word1, word2):
    result = (n_gram_similarity(word1, word2) + cosing_sim(word1, word2)) / 2
    return result




import logging
import os

from bson.json_util import (
    ObjectId
)
from flask import (
    current_app,
    flash,
    request,
    send_file
)
from flask_wtf.csrf import CSRFProtect
from werkzeug.utils import secure_filename

from siena.server.siena_api import blueprint
from siena.core.actions import (
    JSONEncoder,
    allowed_file_nlu,
    allowed_file_knowledge,
    read_yml,
    remove_entry_from_knowledge,
    update_sentences_by_user,
    update_knowledge,
    create_project,
    convert_files,
    get_files,
    get_suggestions,
    get_base_words,
    auto_annotate,
    delete_sentences,
    get_entities_by_project,
    insert_entities_for_project,
    get_projects,
    validate_tempory_knowledge_base,
    is_valid_nlu_yaml,
)
from siena.shared.constants import (
    SIENA_KNOWLEDGE_BASE_PATH,
    LOOKUP_DIR,
    UPLOAD_FOLDER,
    SIENA_IN_PROGRESS_PATH,
)

logger = logging.getLogger(__name__)
csrf = CSRFProtect()


# Save file
@blueprint.route('/fileupload', methods=['POST', 'GET'])
def save_file():
    try:
        logger.debug("File Upload API endpoint was called")
        if request.method == 'POST':
            # check if the post request has the file part
            if 'file' in request.files:
                file = request.files['file']
                # If the user does not select a file, the browser submits an
                # empty file without a filename.
                if file.filename == '':
                    flash('No selected file')
                    return 'file name not found!', 404
                if file and allowed_file_nlu(file.filename):
                    filename = secure_filename(file.filename)
                    path = os.path.join(current_app.config['UPLOAD_FOLDER'], filename)
                    file.save(path)
                    read_yml(path, file.filename)
            else:
                post_data = request.json
                file_path = post_data['FILE_PATH']
                if not file_path:
                    flash('No selected file')
                    return 'file name not found!', 404
                filename = file_path.split("/")[-1]
                read_yml(file_path, filename)
        return {}
    except Exception as e:
        logger.exception(f"Exception occurred in the fileupload endpoint. {e}")
        return {"status": "error", "response": "Error occurred while processing your request"}, 400


# API for sentence management
@blueprint.route('/sentence', methods=['GET', 'PATCH', 'POST', 'DELETE'])
def endpoint_sentences():
    try:
        logger.debug("Sentence API endpoint was called")
        response = {}
        data = []
        if request.method == 'GET':
            with open(SIENA_IN_PROGRESS_PATH, "r", encoding='utf-8') as file:
                line_id = 0
                while line := file.readline().rstrip():
                    row = {}
                    file_line = line.split('<sep>')
                    row["sentence"] = file_line[1]
                    row["intent"] = file_line[0]
                    row["id"] = line_id
                    line_id += 1
                    data.append(row)
                response["data"] = data
            return JSONEncoder().encode(response)
        elif request.method == 'PATCH':
            post_data = request.json
            sentences = post_data['data']
            for sentence in sentences:
                intent = sentence['INTENT']
                entity = sentence['ENTITY']
                mode = ""
                if "MODE" in sentence.keys():
                    mode = sentence['MODE']
                highlighted = sentence['HIGHLIGHTED']
                sentence_id = int(sentence['id'])
                text = sentence['TEXT']
                if mode == "ENTITY_REMOVE":
                    remove_entry_from_knowledge(entity.strip(), highlighted.strip())
                else:
                    if intent != "" and entity != "":
                        update_knowledge(entity.strip(), highlighted.strip())

                update_sentences_by_user(sentence_id, text, intent)
            return response
        elif request.method == 'DELETE':
            post_data = request.json
            sentences = post_data['data']
            for sentence in sentences:
                sentence_id = ObjectId(sentence['_id'])
                delete_sentences(sentence_id)
            return response
        else:
            flash('Invalid request')
            return response
    except Exception as e:
        logger.exception(f"Exception occurred in the sentence endpoint. {e}")
        return {"status": "error", "response": "Error occurred while processing your request"}, 400


#  API for entity management
@blueprint.route('/entity', methods=['GET', 'POST'])
def endpoint_entity():
    try:
        logger.debug("Entity API endpoint was called")
        data = {}
        if request.method == 'GET':
            data["ENTITIES"] = get_entities_by_project()
            return data
        elif request.method == 'POST':
            post_data = request.json
            entities = post_data['data']
            insert_entities_for_project(entities)
            data['messege'] = "Updated"
            return data
        else:
            return 'bad request!', 400
    except Exception as e:
        logger.exception(f"Exception occurred in the entity endpoint. {e}")
        return {"status": "error", "response": "Error occurred while processing your request"}, 400


#  API for project management
@blueprint.route('/project', methods=['GET', 'POST'])
def endpoint_project():
    try:
        logger.debug("Project API endpoint was called")
        data = {}
        if request.method == 'GET':
            user_id = request.cookies.get('userId', None)
            if not user_id:
                return 'user not found!', 404
            else:
                data["PROJECTS"] = get_projects(user_id)
                return JSONEncoder().encode(data)
        if request.method == 'POST':
            user_id = request.cookies.get('userId', None)
            post_data = request.json
            project_name = post_data['NAME']
            project_type = post_data['TYPE']
            if not user_id and not project_name and not project_type:
                return {"status": "error", "response": "data not found!"}, 404
            else:
                data["PROJECT_ID"] = create_project(user_id, project_name, project_type)
                return JSONEncoder().encode(data)
        else:
            return {"status": "error", "response": "bad request!"}, 400
    except Exception as e:
        logger.exception(f"Exception occurred in the project endpoint. {e}")
        return {"status": "error", "response": "Error occurred while processing your request"}, 400


#  API for file management
@blueprint.route('/file', methods=['GET', 'POST'])
def endpoint_file():
    try:
        logger.debug("File API endpoint was called")
        data = {}
        if request.method == 'GET':
            data["FILES"] = get_files()
            return JSONEncoder().encode(data)
        else:
            return {"status": "error", "response": "bad request!"}, 400
    except Exception as e:
        logger.exception(f"Exception occurred in the file endpoint. {e}")
        return {"status": "error", "response": "Error occurred while processing your request"}, 400


@blueprint.route('/save', methods=['GET', 'POST'])
def endpoint_save():
    try:
        logger.debug("Save API endpoint was called")
        if request.method == 'GET':
            data = convert_files()
            # JSONEncoder().encode(data)
            if data:
                return {"status": "success"}, 200
            else:
                raise Exception()
        else:
            return {"status": "error", "response": "bad request!"}, 400
    except Exception as e:
        logger.exception(f"Exception occurred in /api/siena/save endpoint. {e}")
        return {"status": "error", "response": "Error occurred while processing your request"}, 400


# SIENA knowledge component
@blueprint.route('/knowledge', methods=['POST', 'GET'])
def endpoint_algorithms():
    try:
        logger.debug("Knowledge API endpoint was called")
        data = {}
        if request.method == 'POST':
            post_data = request.json
            text = post_data["TEXT"]
            suggestions = get_suggestions(text)
            if len(suggestions) <1:
                return {"status": "error", "response": "No entities in the entity list"}, 400
            data["SUGGESTIONS"] = suggestions
            return JSONEncoder().encode(data)
        elif request.method == 'GET':
            data["DATA"] = get_base_words()
            return JSONEncoder().encode(data)
        else:
            return {"status": "error", "response": "bad request!"}, 400
    except Exception as e:
        logger.exception(f"Exception occurred in the knowledge endpoint. {e}")
        return {"status": "error", "response": "Error occurred while processing your request"}, 400


# import knowledge
@blueprint.route('/knowledge/upload', methods=['POST', 'GET'])
def upload_knowledge():
    try:
        if request.method == 'POST':
            try:
                # check if the post request has the file part
                if 'file' in request.files:
                    file = request.files['file']
                    # If the user does not select a file, the browser submits an
                    # empty file without a filename.
                    if file.filename == '':
                        flash('No selected file')
                        return {"status": "error", "response": "file name not found!"}, 400
                    if file and allowed_file_knowledge(file.filename):
                        file.save("siena_cache/knowledge_base_temp.csv")
                        if validate_tempory_knowledge_base():
                            return {"status": "success", "response": "knowledge saved"}, 200
                        else:
                            return {"status": "error", "response": "error in file format"}, 400
                    else:
                        return {"status": "error", "response": "error in file format"}, 400

                else:
                    return {"status": "error", "response": "invalid file name!"}, 400
            except Exception as e:
                logger.exception(e)

        else:
            return {"status": "error", "response": "bad request!"}, 400
    except Exception as e:
        logger.exception(f"Exception occurred in the uploading knowledge-base endpoint. {e}")
        return {"status": "error", "response": "Error occurred while processing your request"}, 400


# export knowledge
@blueprint.route('/knowledge/export', methods=['POST', 'GET'])
def export_knowledge():
    try:
        if os.path.isfile(os.path.join(os.getcwd(), SIENA_KNOWLEDGE_BASE_PATH)):
            return send_file(
                path_or_file=os.path.join(os.getcwd(), SIENA_KNOWLEDGE_BASE_PATH),
                as_attachment=True
            )
        else:
            raise Exception()
    except Exception as e:
        logger.exception(f"Exception occurred while exporting the knowledge-base csv. {e}")
        return {"status": "error", "response": "Error occurred while processing your request"}, 400


# SIENA auto annotate algorithm
@blueprint.route('/autoannotate', methods=['POST'])
def endpoint_autoannotate():
    try:
        logger.debug("Auto Annotate API endpoint was called")
        data = {}
        if request.method == 'POST':
            post_data = request.json
            base_word = post_data["BASE_WORD"]
            entity = post_data["ENTITY"]
            is_done = auto_annotate(base_word, entity)
            if is_done:
                return {"status": "success", "response": "auto annotation completed"}, 200
            else:
                return {"status": "error", "response": "auto annotation failed"}, 400
        else:
            return {"status": "error", "response": "bad request!"}, 400
    except Exception as e:
        logger.exception(f"Exception occurred in the autoannotate endpoint. {e}")
        return {"status": "error", "response": "Error occurred while processing your request"}, 400


# file navigation API
@blueprint.route('/navigation', methods=['GET'])
def endpoint_navigation():
    try:
        logger.debug("Navigation API endpoint was called")
        global LOOKUP_DIR
        data = {}
        if request.method == 'GET':
            data["FILES"] = []
            # in bot folder
            for root, dir_names, filenames in os.walk(LOOKUP_DIR):
                for filename in filenames:
                    if filename.endswith(('.YAML', '.YML', '.yaml', '.yml')):
                        files = {}
                        abs_path = os.path.join(root, filename)
                        rel_path = abs_path.replace("\\", "/")
                        if not is_valid_nlu_yaml(rel_path):
                            continue
                        files["PATH"] = rel_path
                        files["NAME"] = filename
                        data["FILES"].append(files)
            # in uploads folder
            for root, dir_names, filenames in os.walk(UPLOAD_FOLDER):
                for filename in filenames:
                    if filename.endswith(('.YAML', '.YML', '.yaml', '.yml')):
                        files = {}
                        abs_path = os.path.join(root, filename)
                        rel_path = abs_path.replace("\\", "/")
                        if not is_valid_nlu_yaml(rel_path):
                            continue
                        files["PATH"] = rel_path
                        files["NAME"] = filename
                        data["FILES"].append(files)

            return JSONEncoder().encode(data)
        else:
            return {"status": "error", "response": "bad request!"}, 400
    except Exception as e:
        logger.exception(f"Exception occurred in the navigation endpoint. {e}")
        return {"status": "error", "response": "Error occurred while processing your request"}, 400



import argparse
import logging
import os
import sys
from typing import NoReturn

from siena.server.siena_server import SIENAServer
from siena.shared.constants import (
    LoggingLevel,
    InterfaceType,
    DEFAULT_SERVER_PORT,
    DEFAULT_SERVER_HOST,
)
from siena.utils.siena_logging_formatter import (
    SIENALoggingFormatter,
    MaxLevelFilter,
)

logger = logging.getLogger()
sys.path.insert(0, os.getcwd())

formatter = SIENALoggingFormatter(format_str='%(asctime)s\t%(levelname)s\t%(name)s - %(message)s')
logging_out = logging.StreamHandler(sys.stdout)
logging_err = logging.StreamHandler(sys.stderr)
logging_out.setFormatter(formatter)
logging_err.setFormatter(formatter)
logging_out.addFilter(MaxLevelFilter(logging.WARNING))
logging_out.setLevel(logging.DEBUG)
logging_err.setLevel(logging.WARNING)
logger.addHandler(logging_out)
logger.addHandler(logging_err)
logger.setLevel(level=logging.INFO)


def create_argument_parser():
    """
    Parses the arguments passed to SIENA through SIENA CLI tool.
    Returns:
        argparse.ArgumentParser()
    """

    parser = argparse.ArgumentParser(
        prog="siena",
        description="starts SIENA CLI"
    )
    subparsers = parser.add_subparsers(
        help='desired SIENA interface to run [cli/server]',
        dest="subparser_name"
    )

    parser_server = subparsers.add_parser(
        name="server",
        help='run SIENA server, a web-based visualization '
             'tool for SIENA.'
    )
    parser_server.add_argument(
        "-p",
        "--port",
        type=int,
        default=None,
        help="the port to start the siena server at.",
    )
    parser_server.add_argument(
        "--debug",
        action="store_true",
        help="sets the logging level to debug mode from "
             "info and flask server debug mode to true.",
    )

    return parser


def _set_logging_level(level: LoggingLevel = LoggingLevel.INFO) -> NoReturn:
    """Sets logging level of SIENA"""

    if level == LoggingLevel.NOTSET:
        logger.setLevel(level=logging.NOTSET)
    elif level == LoggingLevel.DEBUG:
        logger.setLevel(level=logging.DEBUG)
    elif level == LoggingLevel.INFO:
        logger.setLevel(level=logging.INFO)
    elif level == LoggingLevel.WARNING:
        logger.setLevel(level=logging.WARNING)
    elif level == LoggingLevel.ERROR:
        logger.setLevel(level=logging.ERROR)
    elif level == LoggingLevel.CRITICAL:
        logger.setLevel(level=logging.CRITICAL)
    elif level == LoggingLevel.QUIET:
        logging.disable(level=logging.CRITICAL)
    else:
        logger.setLevel(level=logging.INFO)


def run_siena_cli() -> NoReturn:
    logger.info("Running main SIENA CLI.")
    arg_parser = create_argument_parser()
    cmdline_args = arg_parser.parse_args()
    interface = cmdline_args.subparser_name

    if not interface:
        arg_parser.print_help()
        logger.error("Please specify a valid positional arg "
                     "out of [\'server\'] to use SIENA CLI.")
        return

    try:
        if str(interface).lower() == InterfaceType.INTERFACE_SERVER:
            server_port = cmdline_args.port
            debug_mode = cmdline_args.debug
            if debug_mode:
                _set_logging_level(level=LoggingLevel.DEBUG)
            else:
                _set_logging_level(level=LoggingLevel.INFO)

            configs = {
                "server_port": server_port or DEFAULT_SERVER_PORT,
                "server_host": DEFAULT_SERVER_HOST
            }
            siena_server = SIENAServer(
                configs=configs,
                debug_mode=debug_mode,
            )
            siena_server.run()

        else:
            logger.error('One or more incorrect CLI arguments detected. '
                         'Refer "siena -h" to view allowed arguments')
            return
    except KeyboardInterrupt:
        logger.info(f"Gracefully terminating SIENA CLI...")


if __name__ == '__main__':
    logger.error("This script cannot be directly executed. "
                 "please use the 'SIENA CLI' instead.")
    exit(1)



from siena.shared.exceptions.base import SIENABaseException


class SIENAIOException(SIENABaseException):
    pass


class YAMLFormatException(SIENAIOException):
    pass


class NLUFileNotFoundException(SIENAIOException):
    pass


class InvalidFileExtensionException(SIENAIOException):
    pass


class FileSizeInspectingException(SIENAIOException):
    pass



import logging
from setuptools import (
    setup,
    find_packages,
)

from siena.shared.constants import (
    PACKAGE_VERSION,
    PACKAGE_NAME_PYPI,
)

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

requirements = None
long_description = None

try:
    with open("README.md", "r") as fh:
        long_description = fh.read()

    with open("requirements.txt", mode="r", encoding="utf8") as requirements_file:
        requirements = requirements_file.readlines()
    requirements = [str.strip(req) for req in requirements]

except Exception as e:
    long_description = "not provided"
    logger.error(f"couldn't retrieve the long package "
                 f"description or requirements list. {e}")

setup(
    name=PACKAGE_NAME_PYPI,
    version=PACKAGE_VERSION,
    packages=find_packages(),
    include_package_data=True,
    package_data={
        # Include special files needed
        # for init project:
        "": [
            "*.SIENA",
            "*.siena",
            "*.tmp",
            "*.md",
            "*.html",
            "*.css",
            "*.js"
        ],
    },
    description="SIENA tool for efficient "
                "entity annotation.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/siena-nlp/siena",
    author="Akalanka Sakalasooriya",
    author_email="himesha@outlook.com",
    license='Apache License 2.0',
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Framework :: Flask",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    install_requires=[
        "beautifulsoup4==4.11.1",
        "flask>=2.1.2",
        "Flask-Cors==3.0.10",
        "Flask-WTF==1.0.1",
        "nltk==3.6.3",
        "pandas==1.4.3",
        "waitress>=2.1.2",
        "rasa~=2.8.8",
        "ruamel.yaml<0.17.0,>=0.16.5",
    ],
    entry_points={'console_scripts': ['siena = siena.siena_cli:run_siena_cli']}
)



import json
import logging
from hashlib import sha256
from pathlib import Path
from typing import Text, Any
import os
import nltk
import numpy as np
import pandas as pd
from bs4 import BeautifulSoup
from bson.objectid import ObjectId
from nltk.stem.porter import *
from ruamel import yaml
from ruamel.yaml import YAML
from ruamel.yaml.scalarstring import PreservedScalarString

from siena.shared.constants import (
    ALLOWED_EXTENSIONS_KNOWLEDGE,
    ALLOWED_EXTENSIONS_NLU,
    COLUMN_NAME_BASE_WORD,
    COLUMN_NAME_COUNT,
    COLUMN_NAME_ENTITY,
    NLTK_WORDNET,
    NLTK_PUNKT,
    SIENA_KNOWLEDGE_BASE_PATH,
    SIENA_IN_PROGRESS_PATH,
    FilePermission,
    Encoding,
    SEP_TAG,
    NEW_LINE_TAG,
    SIENA_ENTITIES_PATH,
    SIENA_CONFIG_PATH,
    UNDEFINED_TAG,
    DEFAULT_NLU_YAML_TAG,
    DEFAULT_NLU_INTENT_TAG,
    DEFAULT_NLU_EXAMPLES_TAG,
    DEFAULT_NLU_YAML_VERSION,
    SIENA_TEMP_KNOWLEDGE_BASE_PATH
)

from siena.core.similarity import (
    base_form_convetor,
    generate_ngrams_sent,
    similarity,
)
from siena.utils.io import file_exists
logger = logging.getLogger(__name__)
en_stemmer = PorterStemmer()
nltk.download(NLTK_WORDNET, quiet=True)
nltk.download(NLTK_PUNKT, quiet=True)
yml = YAML()
yml.indent(mapping=2, sequence=4, offset=2)
yml.preserve_quotes = True
yml.explicit_start = False

knowledge = pd.DataFrame(columns=[COLUMN_NAME_BASE_WORD, COLUMN_NAME_ENTITY, COLUMN_NAME_COUNT])
siena_kb = Path(SIENA_KNOWLEDGE_BASE_PATH)
if siena_kb.is_file():
    knowledge = pd.read_csv(SIENA_KNOWLEDGE_BASE_PATH, index_col=0)
    knowledge.drop(knowledge[knowledge[COLUMN_NAME_COUNT] < 1].index, inplace=True)


class JSONEncoder(json.JSONEncoder):
    def default(self, o) -> Text:
        if isinstance(o, ObjectId):
            return str(o)
        return json.JSONEncoder.default(self, o)


def allowed_file_nlu(filename: Text) -> bool:
    r"""Checks for yaml file extension validation
    File name - name of the file want to check
    """
    return '.' in filename and \
           filename.rsplit('.', 1)[1].lower() in ALLOWED_EXTENSIONS_NLU


def allowed_file_knowledge(filename: Text) -> Any:
    return '.' in filename and \
           filename.rsplit('.', 1)[1].lower() in ALLOWED_EXTENSIONS_KNOWLEDGE


def get_sentences(document_id, range_start, range_end) -> bool:
    logger.debug(f"get_sentences was called. (doc id: "
                 f"{document_id} start: {range_start} "
                 f"end: {range_end}")
    return True


def update_sentences_by_user(line_id, text, intent):
    with open(SIENA_IN_PROGRESS_PATH, FilePermission.READ, encoding=Encoding.UTF8) as file:
        try:
            document = file.readlines()
            document[line_id] = intent + SEP_TAG + text + NEW_LINE_TAG
        except Exception as e:
            logger.exception(f"Exception occurred. {e}")
            return False

    with open(SIENA_IN_PROGRESS_PATH, FilePermission.WRITE_PLUS, encoding=Encoding.UTF8) as file:
        try:
            file.writelines(document)
        except Exception as e:
            logger.exception(f"Exception occurred. {e}")
            return False
    return True


def delete_sentences(sentence_id) -> bool:
    logger.debug(f"delete_sentences was called. sentence id: {sentence_id}")
    return True


def read_yml(path_to_file, file_name) -> None:
    with open(SIENA_CONFIG_PATH, FilePermission.WRITE, encoding=Encoding.UTF8) as file:
        try:
            file.write(f"file_name={file_name}\nfile_path={path_to_file}\n")
        except Exception as e:
            logger.exception(f"Exception occurred. {e}")
    data = {}
    with open(path_to_file, FilePermission.READ, encoding=Encoding.UTF8) as stream:
        try:
            data = yaml.safe_load(stream)
            f = open(SIENA_IN_PROGRESS_PATH, FilePermission.WRITE_PLUS)
            f.write("")
            f.close()
        except yaml.YAMLError as e:
            logger.exception(f"Exception occurred. {e}")

    intents = data[DEFAULT_NLU_YAML_TAG]
    for single_intent in intents:
        intent = single_intent[DEFAULT_NLU_INTENT_TAG]
        examples = single_intent[DEFAULT_NLU_EXAMPLES_TAG].split(NEW_LINE_TAG)
        sentences = [i[2:] for i in examples]
        sentences = list(filter(lambda a: a != '', sentences))
        for sentence in sentences:
            line = intent + SEP_TAG + extract_entities(sentence)
            line = line.replace(NEW_LINE_TAG, '')
            with open(SIENA_IN_PROGRESS_PATH, FilePermission.APPEND_PLUS, encoding=Encoding.UTF8) as f:
                f.write(line + NEW_LINE_TAG)


def get_entities_by_project():
    entities = []
    # entity:value<sep>#color
    try:
        with open(SIENA_ENTITIES_PATH, FilePermission.READ, encoding=Encoding.UTF8) as file:
            while line := file.readline().rstrip():
                row = {}
                file_line = line.split(SEP_TAG)
                entity_name, entity_value = file_line[0].split(":")
                row["ENTITY_NAME"] = entity_name
                row["ENTITY_REPLACER"] = entity_value
                row["ENTITY_COLOR"] = file_line[1]
                entities.append(row)
    except Exception as e:
        logger.exception(f"Exception occurred. {e}")
    return entities


def update_knowledge(entity, highlighted_text):
    global knowledge
    base_word = base_form_convetor(highlighted_text)
    new_row = pd.DataFrame([[base_word, entity, 1]],
                           columns=[COLUMN_NAME_BASE_WORD, COLUMN_NAME_ENTITY, COLUMN_NAME_COUNT])
    knowledge = pd.concat([knowledge, new_row])
    knowledge = pd.DataFrame({COLUMN_NAME_COUNT: knowledge.groupby([COLUMN_NAME_BASE_WORD, COLUMN_NAME_ENTITY])[
        COLUMN_NAME_COUNT].sum()}).reset_index()
    knowledge.to_csv(SIENA_KNOWLEDGE_BASE_PATH)
    return True


def remove_entry_from_knowledge(entity, highlighted_text) -> bool:
    r"""Remove entry from knowledge
    entity - name of the file entity
    """
    global knowledge
    base_word = base_form_convetor(highlighted_text)
    locations_of_matching_base_word = knowledge.index[
        (knowledge[COLUMN_NAME_BASE_WORD] == base_word) & (knowledge[COLUMN_NAME_ENTITY] == entity)].tolist()
    for index in locations_of_matching_base_word:
        pointer = index - 1
        if knowledge.iloc[pointer, 2] > 0:
            knowledge.iloc[pointer, 2] = knowledge.iloc[pointer, 2] - 1
    knowledge.drop(knowledge[knowledge[COLUMN_NAME_COUNT] < 1].index, inplace=True)
    knowledge.to_csv(SIENA_KNOWLEDGE_BASE_PATH)
    return True


def get_suggestions(text: str):
    r"""Get suggestions for annotation
    text - text that want suggestions
    """
    text = text.strip()
    global knowledge
    entities = []
    try:
        with open(SIENA_ENTITIES_PATH, FilePermission.READ, encoding=Encoding.UTF8) as file:
            while line := file.readline().rstrip():
                file_line = line.split(SEP_TAG)
                row = file_line[0]  # name of entity
                entities.append(row)

        if len(entities) < 1:
            suggestions = []
            return suggestions
        entities_df = pd.DataFrame([])
        entities_df["default_entities"] = entities
        idx = knowledge.groupby([COLUMN_NAME_BASE_WORD])[COLUMN_NAME_COUNT].transform(max) == knowledge[
            COLUMN_NAME_COUNT]
        knowledge_filtered = knowledge[idx]

        stemmed_text = base_form_convetor(text)
        filtered_base_forms = list(knowledge_filtered[COLUMN_NAME_BASE_WORD])
        similarity_scores_of_base_words = []
        for base_word in filtered_base_forms:
            similarity_scores_of_base_words.append(similarity(stemmed_text, base_word))
        knowledge_filtered['similarity'] = similarity_scores_of_base_words
        # group by entity type
        knowledge_filtered.reset_index(drop=True, inplace=True)
        idx = knowledge_filtered.groupby([COLUMN_NAME_ENTITY])['similarity'].transform(max) == knowledge_filtered[
            'similarity']
        knowledge_filtered = knowledge_filtered[idx]
        knowledge_filtered_merged = pd.merge(
            entities_df,
            knowledge_filtered,
            left_on='default_entities',
            right_on=COLUMN_NAME_ENTITY, how='left'
        )
        knowledge_filtered_merged = knowledge_filtered_merged[["default_entities", "similarity"]]
        knowledge_filtered_merged_nona = knowledge_filtered_merged.dropna()
        knowledge_filtered_merged_na = knowledge_filtered_merged.loc[knowledge_filtered_merged['similarity'].isna()]
        nan_keys = list(knowledge_filtered_merged_na["default_entities"])
        nan_similarity = []
        for word in nan_keys:
            sim = similarity(text, word)
            nan_similarity.append(sim)
        knowledge_filtered_merged_na.loc[:, "similarity"] = nan_similarity

        # sorting
        knowledge_filtered_merged_nona = knowledge_filtered_merged_nona.sort_values('similarity', ascending=False)
        knowledge_filtered_merged_na = knowledge_filtered_merged_na.sort_values('similarity', ascending=False)
        knowledge_filtered_merged = pd.concat(
            [knowledge_filtered_merged_nona, knowledge_filtered_merged_na],
            ignore_index=True
        )
        suggestions = list(knowledge_filtered_merged["default_entities"])
        return suggestions
    except Exception as e:
        logger.exception(f"Exception occurred. {e}")
        return list()


def get_projects(user_id) -> bool:
    logger.debug(f"get_projects was called. user id: {user_id}")
    return True


def init_project(uploads: bool = True, exports: bool = True, cache: bool = True) -> bool:
    r"""Creates mandatory file and folders for SIENA
    None
    """
    if uploads:
        Path("uploads").mkdir(parents=True, exist_ok=True)
    if exports:
        Path("exports").mkdir(parents=True, exist_ok=True)
    if cache:
        Path("siena_cache").mkdir(parents=True, exist_ok=True)

    fle = Path(SIENA_CONFIG_PATH)
    fle.touch(exist_ok=True)

    fle = Path(SIENA_ENTITIES_PATH)
    fle.touch(exist_ok=True)
    fle = Path(SIENA_IN_PROGRESS_PATH)
    fle.touch(exist_ok=True)

    return True


def auto_annotate(base_form: str, entity: str) -> bool:
    global knowledge
    auto_annotate_count = 0

    # validate base_form
    if base_form not in list(knowledge[COLUMN_NAME_BASE_WORD]):
        return False
    base_form_token = base_form.split(" ")
    n = len(base_form_token)

    # loop start
    with open(SIENA_IN_PROGRESS_PATH, FilePermission.READ, encoding=Encoding.UTF8) as file:
        try:
            in_progress_text = file.readlines()
        except Exception as e:
            logger.exception(f"Exception occurred. {e}")
            return False
    if len(in_progress_text) <= 0:
        return False

    auto_annotated_text = []
    try:
        for line in in_progress_text:
            line_split = line.split(SEP_TAG)
            word_line = " ".join(str(line_split[1]).split())
            intent = " ".join(str(line_split[0]).split())

            soup = BeautifulSoup(word_line, 'html.parser')
            word_line = str(soup)
            div_list = soup.find_all("div", {"name": "highlighted"})
            key_list_ = {}

            for single_div in div_list:
                key_sha256 = sha256(single_div.encode(Encoding.UTF8)).hexdigest()
                key_list_[key_sha256] = str(single_div)
                word_line = word_line.replace(str(single_div), f' {key_sha256} ')

            word_line = " ".join(word_line.split())
            n_grams = generate_ngrams_sent(word_line, n)
            n_grams_len = len(n_grams) - 1
            re_word = ""
            auto_annotate_mapper = {}

            for index, element in enumerate(n_grams):
                if base_form == base_form_convetor(element):
                    auto_annotate_count += 1
                    tag = f"<div class='card-highlighted-text' name='highlighted' data='{entity}'>" \
                          f"{element}<span class='card-highlighted-text-close' style='visibility: hidden;'>" \
                          f"<i class='ms-Icon ms-Icon--ChromeClose ms-fontColor-white'></i></span> </div>"

                    sha_key = sha256(tag.encode(Encoding.UTF8)).hexdigest()
                    auto_annotate_mapper[sha_key] = tag
                    # replace element with sha key
                    element = sha_key
                if index == n_grams_len:
                    # last element
                    re_word = f"{re_word} {element}"
                else:
                    re_word = f"{re_word} {element.split()[0]}"

            # re-constructed word line with hash
            re_word = re_word.strip()
            re_word = " ".join(re_word.split())
            for hash_key, tag in key_list_.items():
                re_word = re_word.replace(hash_key, tag)
            for hash_key, tag in auto_annotate_mapper.items():
                re_word = re_word.replace(hash_key, tag)
            #
            auto_annotated_line = f"{intent}<sep>{re_word}"
            auto_annotated_text.append(auto_annotated_line)

        # update knowledge
        current_count = knowledge[knowledge[COLUMN_NAME_BASE_WORD] == base_form]
        current_count = int(current_count.iloc[0, 2])
        knowledge.loc[knowledge[COLUMN_NAME_BASE_WORD] == base_form, COLUMN_NAME_COUNT] = (
                current_count + auto_annotate_count)
        knowledge.to_csv(SIENA_KNOWLEDGE_BASE_PATH)

        # write to file
        # reset in-progress file
        with open(SIENA_IN_PROGRESS_PATH, FilePermission.WRITE_PLUS, encoding=Encoding.UTF8) as f:
            f.write("")

        # append to file
        for line in auto_annotated_text:
            line = line.replace(NEW_LINE_TAG, '')
            with open(SIENA_IN_PROGRESS_PATH, "a+", encoding=Encoding.UTF8) as f:
                f.write(line + NEW_LINE_TAG)
    except Exception as e:
        logger.error(f"Exception occurred in SIENA actions. More Info: {e}")
        return False
    return True


def create_project(user_id, project_name, project_type):
    logger.debug(f"create_project was called. uid: {user_id} proj name: {project_name} type: {project_type}")
    return True


def get_base_words() -> Any:
    global knowledge
    data = knowledge.drop(knowledge[knowledge[COLUMN_NAME_COUNT] < 1].index)
    data = data[[COLUMN_NAME_BASE_WORD, COLUMN_NAME_ENTITY]].to_dict('records')
    return data


def create_document(project_id, name) -> bool:
    logger.debug(f"create_doc was called. proj id: {project_id} name: {name}")
    return True


def insert_entities_for_project(entities):
    with open(SIENA_ENTITIES_PATH, FilePermission.WRITE, encoding=Encoding.UTF8) as file:
        text = ""
        try:
            for line in entities:
                row = line['ENTITY_NAME'] + \
                      ":" + \
                      line['ENTITY_REPLACER'] + \
                      SEP_TAG + line['ENTITY_COLOR'] + \
                      NEW_LINE_TAG
                if row not in text:
                    text += row
            file.write(text)
        except Exception as e:
            logger.exception(f"Exception occurred. {e}")
            return False
    return True


def get_files():
    data = []
    name = UNDEFINED_TAG
    with open(SIENA_CONFIG_PATH, FilePermission.READ, encoding=Encoding.UTF8) as file:
        while line := file.readline().rstrip():
            key_, value_ = line.split('=')
            if key_ == "file_name":
                name = value_
            else:
                continue
    data.append({'NAME': name})
    return data


def get_file_path():
    path = ""
    with open(SIENA_CONFIG_PATH, FilePermission.READ, encoding=Encoding.UTF8) as file:
        while line := file.readline().rstrip():
            key_, value_ = line.split('=')
            if key_ == "file_path":
                path = value_
            else:
                continue
    return path


def str_presenter(dumper, data):
    if len(data.splitlines()) > 1:  # check for multiline string
        return dumper.represent_scalar('tag:yaml.org,2002:str', data, style='|')
    return dumper.represent_scalar('tag:yaml.org,2002:str', data)


def save_data_to_domain_file() -> None:
    # ./domain.yml
    domain_file = "./domain.yml"
    data = {}
    with open(domain_file, FilePermission.READ, encoding=Encoding.UTF8) as stream:
        try:
            data = yaml.safe_load(stream)
        except Exception as e:
            logger.exception(f"Exception occurred. {e}")

    # in data obj => 'entities', 'slots'
    entities_text = ""
    with open(SIENA_ENTITIES_PATH, FilePermission.READ, encoding=Encoding.UTF8) as file:
        try:
            entities_text = file.read()
        except Exception as e:
            logger.exception(f"Exception occurred. {e}")
            pass
    entities_text = entities_text.strip()
    entity_lines = entities_text.split(NEW_LINE_TAG)

    slots = {}
    for line in entity_lines:
        entity_, color_ = line.split(SEP_TAG)
        entity, value_ = entity_.split(":")
        #
        slot_item = {}
        if entity not in slots.keys():
            slots[entity] = {}
            slot_item["type"] = "categorical"
            slot_item["values"] = [value_]
            slots[entity] = slot_item
        else:
            slot_item = slots[entity]
            slot_item["values"].append(value_)
            slots[entity] = slot_item

    data["slots"] = slots
    data["entities"] = list(slots.keys())
    with open(domain_file, FilePermission.WRITE_PLUS, encoding=Encoding.UTF8) as file:
        yml.dump(data, file)
        file.close()


def convert_files():
    try:
        data = {"version": DEFAULT_NLU_YAML_VERSION, "nlu": []}
        processed_yaml = {}
        entities_mapper = {}
        with open(SIENA_ENTITIES_PATH, FilePermission.READ, encoding=Encoding.UTF8) as file:
            try:
                text_data = file.readlines()
                for line in text_data:
                    entity, replacer = line.split(SEP_TAG)[0].split(":")
                    entities_mapper[entity] = replacer

            except Exception as e:
                logger.exception(f"Exception occurred. {e}")
                return False

        with open(SIENA_IN_PROGRESS_PATH, FilePermission.READ, encoding=Encoding.UTF8) as file:
            try:
                in_progress_text = file.readlines()
            except Exception as e:
                logger.exception(f"Exception occurred. {e}")
                return False

        for document_line in in_progress_text:
            intent, text = document_line.split(SEP_TAG)
            text = text.strip()
            intent = intent.strip()
            soup = BeautifulSoup(text, 'html.parser')
            input_tag = soup.find_all("div", {"name": "highlighted"})
            sentence = ""
            if len(input_tag) > 0:
                for tag in input_tag:
                    try:
                        line = str(soup)
                        attribute = tag.get('data')
                        entity_ = str(attribute).split(":")[0]
                        value_ = str(attribute).split(":")[1]
                        content = str(tag.text).strip()
                        # value_ = entities_mapper[attribute]  # replacer
                        # attribute = attribute.replace(f"{value_}:","")
                        data_values = f'"entity":"{entity_}","value":"{value_}"'
                        section = f'[{content}]{{{data_values}}} '
                        line = line.replace(str(tag), section, 1)
                        sentence = line
                    except Exception as e:
                        logger.exception(f"Exception occurred. {e}")
                        continue
            else:
                sentence = text

            sentence = " ".join(sentence.split())
            if intent in processed_yaml.keys():
                processed_yaml[intent] += f"- {sentence}\n"
            else:
                processed_yaml[intent] = f"- {sentence}\n"

        for key_, value_ in processed_yaml.items():
            section = {"intent": key_, "examples": PreservedScalarString(value_)}
            data["nlu"].append(section)

        file_path = get_file_path()

        with open(file_path, FilePermission.WRITE_PLUS, encoding=Encoding.UTF8) as file:
            yml.dump(data, file)
            file.close()

        # update domain file
        save_data_to_domain_file()

        return True
    except Exception as e:
        logger.exception(f"Exception occurred while saving NLU files. {e}")
        return False


def extract_entities(word):
    return_txt = word
    conjunction_points = [m.start() for m in re.finditer("]{", word)]
    for point in conjunction_points:
        annotated_word = ""
        entity = ""
        start = 0
        end = -1
        # find word -> backwards
        backward = list(range(point, -1, -1))
        for position in backward:
            if word[position] == "[":
                annotated_word = word[position + 1:point]
                start = position
                break

        # find entity -> forward
        forward = list(range(point, len(word), +1))
        for position in forward:
            if word[position] == "}":
                entity_obj = word[point + 1:position + 1]
                res = json.loads(entity_obj)
                entity = res["entity"]
                end = position + 1
                break
        template = f'<div class="card-highlighted-text" name="highlighted" data="{entity}">' \
                   f'{annotated_word}<span class="card-highlighted-text-close" style="visibility: ' \
                   f'hidden;"><i class="ms-Icon ms-Icon--ChromeClose ms-fontColor-white"></i></span>' \
                   f'</div>'
        phrase = word[start:end]
        return_txt = return_txt.replace(phrase, template)
    return return_txt

def validate_tempory_knowledge_base() -> bool:
    global knowledge
    try:
        columns = np.array([COLUMN_NAME_BASE_WORD, COLUMN_NAME_ENTITY, COLUMN_NAME_COUNT])
        knowledge_temp = pd.read_csv(SIENA_TEMP_KNOWLEDGE_BASE_PATH, index_col=0)
        if knowledge_temp.shape[0] <= 0:
            # Gives number of rows
            return False

        if knowledge.shape[1] != knowledge_temp.shape[1]:
            # Gives number of columns
            return False

        columns_temp = np.array(knowledge_temp.columns.values.tolist())
        result = list(columns_temp == columns)
        if False in result:
            return False
        else:
            if file_exists(SIENA_KNOWLEDGE_BASE_PATH):
                os.remove(SIENA_KNOWLEDGE_BASE_PATH)

            os.rename(SIENA_TEMP_KNOWLEDGE_BASE_PATH, SIENA_KNOWLEDGE_BASE_PATH)
            knowledge = knowledge_temp
            return True

    except Exception as e:
        logger.exception(f"Exception occurred. {e}")
        return False


def get_knowledge_base_text():
    with open(SIENA_KNOWLEDGE_BASE_PATH, FilePermission.READ, encoding=Encoding.UTF8) as file:
        try:
            knowledge_base = file.readlines()
            return knowledge_base
        except Exception as e:
            logger.exception(f"Exception occurred. {e}")
            return False


def is_valid_nlu_yaml(path_to_file):
    with open(path_to_file, FilePermission.READ, encoding=Encoding.UTF8) as stream:
        try:
            data = yaml.safe_load(stream)
            if DEFAULT_NLU_YAML_TAG in data.keys():
                if data[DEFAULT_NLU_YAML_TAG] != None:
                    return True
        except yaml.YAMLError as e:
            logger.exception(f"Exception occurred. {e}")
            return False
    return False